"""
Utilities to aid testing.


"""
from __future__ import print_function

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import hashlib
from yt.extern.six import string_types
from yt.extern.six.moves import cPickle
import itertools as it
import numpy as np
import functools
import importlib
import os
import unittest
from yt.funcs import iterable
from yt.config import ytcfg
# we import this in a weird way from numpy.testing to avoid triggering
# flake8 errors from the unused imports. These test functions are imported
# elsewhere in yt from here so we want them to be imported here.
from numpy.testing import assert_array_equal, assert_almost_equal  # NOQA
from numpy.testing import assert_approx_equal, assert_array_almost_equal  # NOQA
from numpy.testing import assert_equal, assert_array_less  # NOQA
from numpy.testing import assert_string_equal  # NOQA
from numpy.testing import assert_array_almost_equal_nulp  # NOQA
from numpy.testing import assert_allclose, assert_raises  # NOQA
from numpy.random import RandomState
from yt.convenience import load
from yt.units.yt_array import YTArray, YTQuantity
from yt.utilities.exceptions import YTUnitOperationError

# Expose assert_true and assert_less_equal from unittest.TestCase
# this is adopted from nose. Doing this here allows us to avoid importing
# nose at the top level.
class _Dummy(unittest.TestCase):
    def nop():
        pass
_t = _Dummy('nop')

assert_true = getattr(_t, 'assertTrue')
assert_less_equal = getattr(_t, 'assertLessEqual')


def assert_rel_equal(a1, a2, decimals, err_msg='', verbose=True):
    # We have nan checks in here because occasionally we have fields that get
    # weighted without non-zero weights.  I'm looking at you, particle fields!
    if isinstance(a1, np.ndarray):
        assert(a1.size == a2.size)
        # Mask out NaNs
        assert((np.isnan(a1) == np.isnan(a2)).all())
        a1[np.isnan(a1)] = 1.0
        a2[np.isnan(a2)] = 1.0
        # Mask out 0
        ind1 = np.array(np.abs(a1) < np.finfo(a1.dtype).eps)
        ind2 = np.array(np.abs(a2) < np.finfo(a2.dtype).eps)
        assert((ind1 == ind2).all())
        a1[ind1] = 1.0
        a2[ind2] = 1.0
    elif np.any(np.isnan(a1)) and np.any(np.isnan(a2)):
        return True
    if not isinstance(a1, np.ndarray) and a1 == a2 == 0.0:
        # NANS!
        a1 = a2 = 1.0
    return assert_almost_equal(np.array(a1)/np.array(a2), 1.0, decimals, err_msg=err_msg,
                               verbose=verbose)

def amrspace(extent, levels=7, cells=8):
    """Creates two numpy arrays representing the left and right bounds of
    an AMR grid as well as an array for the AMR level of each cell.

    Parameters
    ----------
    extent : array-like
        This a sequence of length 2*ndims that is the bounds of each dimension.
        For example, the 2D unit square would be given by [0.0, 1.0, 0.0, 1.0].
        A 3D cylindrical grid may look like [0.0, 2.0, -1.0, 1.0, 0.0, 2*np.pi].
    levels : int or sequence of ints, optional
        This is the number of AMR refinement levels.  If given as a sequence (of
        length ndims), then each dimension will be refined down to this level.
        All values in this array must be the same or zero.  A zero valued dimension
        indicates that this dim should not be refined.  Taking the 3D cylindrical
        example above if we don't want refine theta but want r and z at 5 we would
        set levels=(5, 5, 0).
    cells : int, optional
        This is the number of cells per refinement level.

    Returns
    -------
    left : float ndarray, shape=(npoints, ndims)
        The left AMR grid points.
    right : float ndarray, shape=(npoints, ndims)
        The right AMR grid points.
    level : int ndarray, shape=(npoints,)
        The AMR level for each point.

    Examples
    --------
    >>> l, r, lvl = amrspace([0.0, 2.0, 1.0, 2.0, 0.0, 3.14], levels=(3,3,0), cells=2)
    >>> print l
    [[ 0.     1.     0.   ]
     [ 0.25   1.     0.   ]
     [ 0.     1.125  0.   ]
     [ 0.25   1.125  0.   ]
     [ 0.5    1.     0.   ]
     [ 0.     1.25   0.   ]
     [ 0.5    1.25   0.   ]
     [ 1.     1.     0.   ]
     [ 0.     1.5    0.   ]
     [ 1.     1.5    0.   ]]

    """
    extent = np.asarray(extent, dtype='f8')
    dextent = extent[1::2] - extent[::2]
    ndims = len(dextent)

    if isinstance(levels, int):
        minlvl = maxlvl = levels
        levels = np.array([levels]*ndims, dtype='int32')
    else:
        levels = np.asarray(levels, dtype='int32')
        minlvl = levels.min()
        maxlvl = levels.max()
        if minlvl != maxlvl and (minlvl != 0 or set([minlvl, maxlvl]) != set(levels)):
            raise ValueError("all levels must have the same value or zero.")
    dims_zero = (levels == 0)
    dims_nonzero = ~dims_zero
    ndims_nonzero = dims_nonzero.sum()

    npoints = (cells**ndims_nonzero - 1)*maxlvl + 1
    left = np.empty((npoints, ndims), dtype='float64')
    right = np.empty((npoints, ndims), dtype='float64')
    level = np.empty(npoints, dtype='int32')

    # fill zero dims
    left[:,dims_zero] = extent[::2][dims_zero]
    right[:,dims_zero] = extent[1::2][dims_zero]

    # fill non-zero dims
    dcell = 1.0 / cells
    left_slice =  tuple([slice(extent[2*n], extent[2*n+1], extent[2*n+1]) if \
        dims_zero[n] else slice(0.0,1.0,dcell) for n in range(ndims)])
    right_slice = tuple([slice(extent[2*n+1], extent[2*n], -extent[2*n+1]) if \
        dims_zero[n] else slice(dcell,1.0+dcell,dcell) for n in range(ndims)])
    left_norm_grid = np.reshape(np.mgrid[left_slice].T.flat[ndims:], (-1, ndims))
    lng_zero = left_norm_grid[:,dims_zero]
    lng_nonzero = left_norm_grid[:,dims_nonzero]

    right_norm_grid = np.reshape(np.mgrid[right_slice].T.flat[ndims:], (-1, ndims))
    rng_zero = right_norm_grid[:,dims_zero]
    rng_nonzero = right_norm_grid[:,dims_nonzero]

    level[0] = maxlvl
    left[0,:] = extent[::2]
    right[0,dims_zero] = extent[1::2][dims_zero]
    right[0,dims_nonzero] = (dcell**maxlvl)*dextent[dims_nonzero] + extent[::2][dims_nonzero]
    for i, lvl in enumerate(range(maxlvl, 0, -1)):
        start = (cells**ndims_nonzero - 1)*i + 1
        stop = (cells**ndims_nonzero - 1)*(i+1) + 1
        dsize = dcell**(lvl-1) * dextent[dims_nonzero]
        level[start:stop] = lvl
        left[start:stop,dims_zero] = lng_zero
        left[start:stop,dims_nonzero] = lng_nonzero*dsize + extent[::2][dims_nonzero]
        right[start:stop,dims_zero] = rng_zero
        right[start:stop,dims_nonzero] = rng_nonzero*dsize + extent[::2][dims_nonzero]

    return left, right, level

def fake_random_ds(
        ndims, peak_value = 1.0,
        fields = ("density", "velocity_x", "velocity_y", "velocity_z"),
        units = ('g/cm**3', 'cm/s', 'cm/s', 'cm/s'),
        particle_fields=None, particle_field_units=None,
        negative = False, nprocs = 1, particles = 0, length_unit=1.0,
        unit_system="cgs", bbox=None):
    from yt.frontends.stream.api import load_uniform_grid
    prng = RandomState(0x4d3d3d3)
    if not iterable(ndims):
        ndims = [ndims, ndims, ndims]
    else:
        assert(len(ndims) == 3)
    if not iterable(negative):
        negative = [negative for f in fields]
    assert(len(fields) == len(negative))
    offsets = []
    for n in negative:
        if n:
            offsets.append(0.5)
        else:
            offsets.append(0.0)
    data = {}
    for field, offset, u in zip(fields, offsets, units):
        v = (prng.random_sample(ndims) - offset) * peak_value
        if field[0] == "all":
            data['number_of_particles'] = v.size
            v = v.ravel()
        data[field] = (v, u)
    if particles:
        if particle_fields is not None:
            for field, unit in zip(particle_fields, particle_field_units):
                if field in ('particle_position', 'particle_velocity'):
                    data['io', field] = (
                        prng.random_sample((int(particles), 3)), unit)
                else:
                    data['io', field] = (
                        prng.random_sample(size=int(particles)), unit)
        else:
            for f in ('particle_position_%s' % ax for ax in 'xyz'):
                data['io', f] = (prng.random_sample(size=particles), 'code_length')
            for f in ('particle_velocity_%s' % ax for ax in 'xyz'):
                data['io', f] = (prng.random_sample(size=particles) - 0.5, 'cm/s')
            data['io', 'particle_mass'] = (prng.random_sample(particles), 'g')
        data['number_of_particles'] = particles
    ug = load_uniform_grid(data, ndims, length_unit=length_unit, nprocs=nprocs,
                           unit_system=unit_system, bbox=bbox)
    return ug

_geom_transforms = {
    # These are the bounds we want.  Cartesian we just assume goes 0 .. 1.
    'cartesian'  : ( (0.0, 0.0, 0.0), (1.0, 1.0, 1.0) ),
    'spherical'  : ( (0.0, 0.0, 0.0), (1.0, np.pi, 2*np.pi) ),
    'cylindrical': ( (0.0, 0.0, 0.0), (1.0, 1.0, 2.0*np.pi) ), # rzt
    'polar'      : ( (0.0, 0.0, 0.0), (1.0, 2.0*np.pi, 1.0) ), # rtz
    'geographic' : ( (-90.0, -180.0, 0.0), (90.0, 180.0, 1000.0) ), # latlonalt
    'internal_geographic' :
                   ( (-90.0, -180.0, 0.0), (90.0, 180.0, 1000.0) ), # latlondep
}

def fake_amr_ds(fields = ("Density",), geometry = "cartesian", particles=0):
    from yt.frontends.stream.api import load_amr_grids
    prng = RandomState(0x4d3d3d3)
    LE, RE = _geom_transforms[geometry]
    LE = np.array(LE)
    RE = np.array(RE)
    data = []
    for gspec in _amr_grid_index:
        level, left_edge, right_edge, dims = gspec
        left_edge = left_edge * (RE - LE) + LE
        right_edge = right_edge * (RE - LE) + LE
        gdata = dict(level = level,
                     left_edge = left_edge,
                     right_edge = right_edge,
                     dimensions = dims)
        for f in fields:
            gdata[f] = prng.random_sample(dims)
        if particles:
            for i, f in enumerate('particle_position_%s' % ax for ax in 'xyz'):
                pdata = prng.random_sample(particles)
                pdata /= (right_edge[i] - left_edge[i])
                pdata += left_edge[i]
                gdata['io', f] = (pdata, 'code_length')
            for f in ('particle_velocity_%s' % ax for ax in 'xyz'):
                gdata['io', f] = (prng.random_sample(particles) - 0.5, 'cm/s')
            gdata['io', 'particle_mass'] = (prng.random_sample(particles), 'g')
            gdata['number_of_particles'] = particles
        data.append(gdata)
    bbox = np.array([LE, RE]).T
    return load_amr_grids(data, [32, 32, 32], geometry=geometry, bbox=bbox)

def fake_particle_ds(
        fields = ("particle_position_x",
                  "particle_position_y",
                  "particle_position_z",
                  "particle_mass",
                  "particle_velocity_x",
                  "particle_velocity_y",
                  "particle_velocity_z"),
        units = ('cm', 'cm', 'cm', 'g', 'cm/s', 'cm/s', 'cm/s'),
        negative = (False, False, False, False, True, True, True),
        npart = 16**3, length_unit=1.0):
    from yt.frontends.stream.api import load_particles

    prng = RandomState(0x4d3d3d3)
    if not iterable(negative):
        negative = [negative for f in fields]
    assert(len(fields) == len(negative))
    offsets = []
    for n in negative:
        if n:
            offsets.append(0.5)
        else:
            offsets.append(0.0)
    data = {}
    for field, offset, u in zip(fields, offsets, units):
        if "position" in field:
            v = prng.normal(loc=0.5, scale=0.25, size=npart)
            np.clip(v, 0.0, 1.0, v)
        v = (prng.random_sample(npart) - offset)
        data[field] = (v, u)
    bbox = np.array([[0.0, 1.0], [0.0, 1.0], [0.0, 1.0]])
    ds = load_particles(data, 1.0, bbox=bbox)
    return ds


def fake_tetrahedral_ds():
    from yt.frontends.stream.api import load_unstructured_mesh
    from yt.frontends.stream.sample_data.tetrahedral_mesh import \
        _connectivity, _coordinates

    prng = RandomState(0x4d3d3d3)

    # the distance from the origin
    node_data = {}
    dist = np.sum(_coordinates**2, 1)
    node_data[('connect1', 'test')] = dist[_connectivity]

    # each element gets a random number
    elem_data = {}
    elem_data[('connect1', 'elem')] = prng.rand(_connectivity.shape[0])

    ds = load_unstructured_mesh(_connectivity,
                                _coordinates,
                                node_data=node_data,
                                elem_data=elem_data)
    return ds


def fake_hexahedral_ds():
    from yt.frontends.stream.api import load_unstructured_mesh
    from yt.frontends.stream.sample_data.hexahedral_mesh import \
        _connectivity, _coordinates

    prng = RandomState(0x4d3d3d3)
    # the distance from the origin
    node_data = {}
    dist = np.sum(_coordinates**2, 1)
    node_data[('connect1', 'test')] = dist[_connectivity-1]

    # each element gets a random number
    elem_data = {}
    elem_data[('connect1', 'elem')] = prng.rand(_connectivity.shape[0])

    ds = load_unstructured_mesh(_connectivity-1,
                                _coordinates,
                                node_data=node_data,
                                elem_data=elem_data)
    return ds

def small_fake_hexahedral_ds():
    from yt.frontends.stream.api import load_unstructured_mesh

    _coordinates = np.array([[-1., -1., -1.],
                              [ 0., -1., -1.],
                              [ -0.,  0., -1.],
                              [-1.,  -0., -1.],
                              [-1., -1.,  0.],
                              [ -0., -1.,  0.],
                              [ -0.,  0.,  -0.],
                              [-1.,  0.,  -0.]])
    _connectivity = np.array([[1, 2, 3, 4, 5, 6, 7, 8]])

    # the distance from the origin
    node_data = {}
    dist = np.sum(_coordinates**2, 1)
    node_data[('connect1', 'test')] = dist[_connectivity-1]

    ds = load_unstructured_mesh(_connectivity-1,
                                _coordinates,
                                node_data=node_data)
    return ds



def fake_vr_orientation_test_ds(N = 96, scale=1):
    """
    create a toy dataset that puts a sphere at (0,0,0), a single cube
    on +x, two cubes on +y, and three cubes on +z in a domain from
    [-1*scale,1*scale]**3.  The lower planes
    (x = -1*scale, y = -1*scale, z = -1*scale) are also given non-zero
    values.

    This dataset allows you to easily explore orientations and
    handiness in VR and other renderings

    Parameters
    ----------

    N : integer
       The number of cells along each direction

    scale : float
       A spatial scale, the domain boundaries will be multiplied by scale to
       test datasets that have spatial different scales (e.g. data in CGS units)

    """
    from yt.frontends.stream.api import load_uniform_grid

    xmin = ymin = zmin = -1.0 * scale
    xmax = ymax = zmax = 1.0 * scale

    dcoord = (xmax - xmin)/N

    arr = np.zeros((N,N,N), dtype=np.float64)
    arr[:,:,:] = 1.e-4

    bbox = np.array([ [xmin, xmax], [ymin, ymax], [zmin, zmax] ])

    # coordinates -- in the notation data[i, j, k]
    x = (np.arange(N) + 0.5)*dcoord + xmin
    y = (np.arange(N) + 0.5)*dcoord + ymin
    z = (np.arange(N) + 0.5)*dcoord + zmin

    x3d, y3d, z3d = np.meshgrid(x, y, z, indexing="ij")

    # sphere at the origin
    c = np.array( [0.5*(xmin + xmax), 0.5*(ymin + ymax), 0.5*(zmin + zmax) ] )
    r = np.sqrt((x3d - c[0])**2 + (y3d - c[1])**2 + (z3d - c[2])**2)
    arr[r < 0.05] = 1.0

    arr[abs(x3d - xmin) < 2*dcoord] = 0.3
    arr[abs(y3d - ymin) < 2*dcoord] = 0.3
    arr[abs(z3d - zmin) < 2*dcoord] = 0.3

    # single cube on +x
    xc = 0.75 * scale
    dx = 0.05 * scale
    idx = np.logical_and(np.logical_and(x3d > xc-dx, x3d < xc+dx),
                         np.logical_and(np.logical_and(y3d > -dx, y3d < dx),
                                        np.logical_and(z3d > -dx, z3d < dx)) )
    arr[idx] = 1.0

    # two cubes on +y
    dy = 0.05 * scale
    for yc in [0.65 * scale, 0.85 * scale]:
        idx = np.logical_and(np.logical_and(y3d > yc-dy, y3d < yc+dy),
                             np.logical_and(np.logical_and(x3d > -dy, x3d < dy),
                                            np.logical_and(z3d > -dy, z3d < dy)) )
        arr[idx] = 0.8

    # three cubes on +z
    dz = 0.05 * scale
    for zc in [0.5 * scale, 0.7 * scale, 0.9 * scale]:
        idx = np.logical_and(np.logical_and(z3d > zc-dz, z3d < zc+dz),
                             np.logical_and(np.logical_and(x3d > -dz, x3d < dz),
                                            np.logical_and(y3d > -dz, y3d < dz)) )
        arr[idx] = 0.6

    data = dict(density = (arr, "g/cm**3"))
    ds = load_uniform_grid(data, arr.shape, bbox=bbox)
    return ds

def expand_keywords(keywords, full=False):
    """
    expand_keywords is a means for testing all possible keyword
    arguments in the nosetests.  Simply pass it a dictionary of all the
    keyword arguments and all of the values for these arguments that you
    want to test.

    It will return a list of kwargs dicts containing combinations of
    the various kwarg values you passed it.  These can then be passed
    to the appropriate function in nosetests.

    If full=True, then every possible combination of keywords is produced,
    otherwise, every keyword option is included at least once in the output
    list.  Be careful, by using full=True, you may be in for an exponentially
    larger number of tests!

    Parameters
    ----------

    keywords : dict
        a dictionary where the keys are the keywords for the function,
        and the values of each key are the possible values that this key
        can take in the function

    full : bool
        if set to True, every possible combination of given keywords is
        returned

    Returns
    -------

    array of dicts
        An array of dictionaries to be individually passed to the appropriate
        function matching these kwargs.

    Examples
    --------

    >>> keywords = {}
    >>> keywords['dpi'] = (50, 100, 200)
    >>> keywords['cmap'] = ('arbre', 'kelp')
    >>> list_of_kwargs = expand_keywords(keywords)
    >>> print list_of_kwargs

    array([{'cmap': 'arbre', 'dpi': 50},
           {'cmap': 'kelp', 'dpi': 100},
           {'cmap': 'arbre', 'dpi': 200}], dtype=object)

    >>> list_of_kwargs = expand_keywords(keywords, full=True)
    >>> print list_of_kwargs

    array([{'cmap': 'arbre', 'dpi': 50},
           {'cmap': 'arbre', 'dpi': 100},
           {'cmap': 'arbre', 'dpi': 200},
           {'cmap': 'kelp', 'dpi': 50},
           {'cmap': 'kelp', 'dpi': 100},
           {'cmap': 'kelp', 'dpi': 200}], dtype=object)

    >>> for kwargs in list_of_kwargs:
    ...     write_projection(*args, **kwargs)
    """

    # if we want every possible combination of keywords, use iter magic
    if full:
        keys = sorted(keywords)
        list_of_kwarg_dicts = np.array([dict(zip(keys, prod)) for prod in \
                              it.product(*(keywords[key] for key in keys))])

    # if we just want to probe each keyword, but not necessarily every
    # combination
    else:
        # Determine the maximum number of values any of the keywords has
        num_lists = 0
        for val in keywords.values():
            if isinstance(val, string_types):
                num_lists = max(1.0, num_lists)
            else:
                num_lists = max(len(val), num_lists)

        # Construct array of kwargs dicts, each element of the list is a different
        # **kwargs dict.  each kwargs dict gives a different combination of
        # the possible values of the kwargs

        # initialize array
        list_of_kwarg_dicts = np.array([dict() for x in range(num_lists)])

        # fill in array
        for i in np.arange(num_lists):
            list_of_kwarg_dicts[i] = {}
            for key in keywords.keys():
                # if it's a string, use it (there's only one)
                if isinstance(keywords[key], string_types):
                    list_of_kwarg_dicts[i][key] = keywords[key]
                # if there are more options, use the i'th val
                elif i < len(keywords[key]):
                    list_of_kwarg_dicts[i][key] = keywords[key][i]
                # if there are not more options, use the 0'th val
                else:
                    list_of_kwarg_dicts[i][key] = keywords[key][0]

    return list_of_kwarg_dicts

def requires_module(module):
    """
    Decorator that takes a module name as an argument and tries to import it.
    If the module imports without issue, the function is returned, but if not,
    a null function is returned. This is so tests that depend on certain modules
    being imported will not fail if the module is not installed on the testing
    platform.
    """
    def ffalse(func):
        return lambda: None
    def ftrue(func):
        return func
    try:
        importlib.import_module(module)
    except ImportError:
        return ffalse
    else:
        return ftrue

def requires_file(req_file):
    path = ytcfg.get("yt", "test_data_dir")
    def ffalse(func):
        return lambda: None
    def ftrue(func):
        return func
    if os.path.exists(req_file):
        return ftrue
    else:
        if os.path.exists(os.path.join(path,req_file)):
            return ftrue
        else:
            return ffalse

def disable_dataset_cache(func):
    @functools.wraps(func)
    def newfunc(*args, **kwargs):
        restore_cfg_state = False
        if ytcfg.get("yt", "skip_dataset_cache") == "False":
            ytcfg["yt","skip_dataset_cache"] = "True"
        rv = func(*args, **kwargs)
        if restore_cfg_state:
            ytcfg["yt","skip_dataset_cache"] = "False"
        return rv
    return newfunc

@disable_dataset_cache
def units_override_check(fn):
    units_list = ["length","time","mass","velocity",
                  "magnetic","temperature"]
    ds1 = load(fn)
    units_override = {}
    attrs1 = []
    attrs2 = []
    for u in units_list:
        unit_attr = getattr(ds1, "%s_unit" % u, None)
        if unit_attr is not None:
            attrs1.append(unit_attr)
            units_override["%s_unit" % u] = (unit_attr.v, str(unit_attr.units))
    del ds1
    ds2 = load(fn, units_override=units_override)
    assert(len(ds2.units_override) > 0)
    for u in units_list:
        unit_attr = getattr(ds2, "%s_unit" % u, None)
        if unit_attr is not None:
            attrs2.append(unit_attr)
    assert_equal(attrs1, attrs2)

# This is an export of the 40 grids in IsolatedGalaxy that are of level 4 or
# lower.  It's just designed to give a sample AMR index to deal with.
_amr_grid_index = [
 [ 0,
  [0.0,0.0,0.0],
  [1.0,1.0,1.0],
  [32,32,32],
 ],
 [ 1,
  [0.25,0.21875,0.25],
  [0.5,0.5,0.5],
  [16,18,16],
 ],
 [ 1,
  [0.5,0.21875,0.25],
  [0.75,0.5,0.5],
  [16,18,16],
 ],
 [ 1,
  [0.21875,0.5,0.25],
  [0.5,0.75,0.5],
  [18,16,16],
 ],
 [ 1,
  [0.5,0.5,0.25],
  [0.75,0.75,0.5],
  [16,16,16],
 ],
 [ 1,
  [0.25,0.25,0.5],
  [0.5,0.5,0.75],
  [16,16,16],
 ],
 [ 1,
  [0.5,0.25,0.5],
  [0.75,0.5,0.75],
  [16,16,16],
 ],
 [ 1,
  [0.25,0.5,0.5],
  [0.5,0.75,0.75],
  [16,16,16],
 ],
 [ 1,
  [0.5,0.5,0.5],
  [0.75,0.75,0.75],
  [16,16,16],
 ],
 [ 2,
  [0.5,0.5,0.5],
  [0.71875,0.71875,0.71875],
  [28,28,28],
 ],
 [ 3,
  [0.5,0.5,0.5],
  [0.6640625,0.65625,0.6796875],
  [42,40,46],
 ],
 [ 4,
  [0.5,0.5,0.5],
  [0.59765625,0.6015625,0.6015625],
  [50,52,52],
 ],
 [ 2,
  [0.28125,0.5,0.5],
  [0.5,0.734375,0.71875],
  [28,30,28],
 ],
 [ 3,
  [0.3359375,0.5,0.5],
  [0.5,0.671875,0.6640625],
  [42,44,42],
 ],
 [ 4,
  [0.40625,0.5,0.5],
  [0.5,0.59765625,0.59765625],
  [48,50,50],
 ],
 [ 2,
  [0.5,0.28125,0.5],
  [0.71875,0.5,0.71875],
  [28,28,28],
 ],
 [ 3,
  [0.5,0.3359375,0.5],
  [0.671875,0.5,0.6640625],
  [44,42,42],
 ],
 [ 4,
  [0.5,0.40625,0.5],
  [0.6015625,0.5,0.59765625],
  [52,48,50],
 ],
 [ 2,
  [0.28125,0.28125,0.5],
  [0.5,0.5,0.71875],
  [28,28,28],
 ],
 [ 3,
  [0.3359375,0.3359375,0.5],
  [0.5,0.5,0.671875],
  [42,42,44],
 ],
 [ 4,
  [0.46484375,0.37890625,0.50390625],
  [0.4765625,0.390625,0.515625],
  [6,6,6],
 ],
 [ 4,
  [0.40625,0.40625,0.5],
  [0.5,0.5,0.59765625],
  [48,48,50],
 ],
 [ 2,
  [0.5,0.5,0.28125],
  [0.71875,0.71875,0.5],
  [28,28,28],
 ],
 [ 3,
  [0.5,0.5,0.3359375],
  [0.6796875,0.6953125,0.5],
  [46,50,42],
 ],
 [ 4,
  [0.5,0.5,0.40234375],
  [0.59375,0.6015625,0.5],
  [48,52,50],
 ],
 [ 2,
  [0.265625,0.5,0.28125],
  [0.5,0.71875,0.5],
  [30,28,28],
 ],
 [ 3,
  [0.3359375,0.5,0.328125],
  [0.5,0.65625,0.5],
  [42,40,44],
 ],
 [ 4,
  [0.40234375,0.5,0.40625],
  [0.5,0.60546875,0.5],
  [50,54,48],
 ],
 [ 2,
  [0.5,0.265625,0.28125],
  [0.71875,0.5,0.5],
  [28,30,28],
 ],
 [ 3,
  [0.5,0.3203125,0.328125],
  [0.6640625,0.5,0.5],
  [42,46,44],
 ],
 [ 4,
  [0.5,0.3984375,0.40625],
  [0.546875,0.5,0.5],
  [24,52,48],
 ],
 [ 4,
  [0.546875,0.41796875,0.4453125],
  [0.5625,0.4375,0.5],
  [8,10,28],
 ],
 [ 4,
  [0.546875,0.453125,0.41796875],
  [0.5546875,0.48046875,0.4375],
  [4,14,10],
 ],
 [ 4,
  [0.546875,0.4375,0.4375],
  [0.609375,0.5,0.5],
  [32,32,32],
 ],
 [ 4,
  [0.546875,0.4921875,0.41796875],
  [0.56640625,0.5,0.4375],
  [10,4,10],
 ],
 [ 4,
  [0.546875,0.48046875,0.41796875],
  [0.5703125,0.4921875,0.4375],
  [12,6,10],
 ],
 [ 4,
  [0.55859375,0.46875,0.43359375],
  [0.5703125,0.48046875,0.4375],
  [6,6,2],
 ],
 [ 2,
  [0.265625,0.28125,0.28125],
  [0.5,0.5,0.5],
  [30,28,28],
 ],
 [ 3,
  [0.328125,0.3359375,0.328125],
  [0.5,0.5,0.5],
  [44,42,44],
 ],
 [ 4,
  [0.4140625,0.40625,0.40625],
  [0.5,0.5,0.5],
  [44,48,48],
 ],
]

def check_results(func):
    r"""This is a decorator for a function to verify that the (numpy ndarray)
    result of a function is what it should be.

    This function is designed to be used for very light answer testing.
    Essentially, it wraps around a larger function that returns a numpy array,
    and that has results that should not change.  It is not necessarily used
    inside the testing scripts themselves, but inside testing scripts written
    by developers during the testing of pull requests and new functionality.
    If a hash is specified, it "wins" and the others are ignored.  Otherwise,
    tolerance is 1e-8 (just above single precision.)

    The correct results will be stored if the command line contains
    --answer-reference , and otherwise it will compare against the results on
    disk.  The filename will be func_results_ref_FUNCNAME.cpkl where FUNCNAME
    is the name of the function being tested.

    If you would like more control over the name of the pickle file the results
    are stored in, you can pass the result_basename keyword argument to the
    function you are testing.  The check_results decorator will use the value
    of the keyword to construct the filename of the results data file.  If
    result_basename is not specified, the name of the testing function is used.

    This will raise an exception if the results are not correct.

    Examples
    --------

    >>> @check_results
    ... def my_func(ds):
    ...     return ds.domain_width

    >>> my_func(ds)

    >>> @check_results
    ... def field_checker(dd, field_name):
    ...     return dd[field_name]

    >>> field_checker(ds.all_data(), 'density', result_basename='density')

    """
    def compute_results(func):
        def _func(*args, **kwargs):
            name = kwargs.pop("result_basename", func.__name__)
            rv = func(*args, **kwargs)
            if hasattr(rv, "convert_to_base"):
                rv.convert_to_base()
                _rv = rv.ndarray_view()
            else:
                _rv = rv
            mi = _rv.min()
            ma = _rv.max()
            st = _rv.std(dtype="float64")
            su = _rv.sum(dtype="float64")
            si = _rv.size
            ha = hashlib.md5(_rv.tostring()).hexdigest()
            fn = "func_results_ref_%s.cpkl" % (name)
            with open(fn, "wb") as f:
                cPickle.dump( (mi, ma, st, su, si, ha), f)
            return rv
        return _func
    from yt.mods import unparsed_args
    if "--answer-reference" in unparsed_args:
        return compute_results(func)

    def compare_results(func):
        def _func(*args, **kwargs):
            name = kwargs.pop("result_basename", func.__name__)
            rv = func(*args, **kwargs)
            if hasattr(rv, "convert_to_base"):
                rv.convert_to_base()
                _rv = rv.ndarray_view()
            else:
                _rv = rv
            vals = (_rv.min(),
                    _rv.max(),
                    _rv.std(dtype="float64"),
                    _rv.sum(dtype="float64"),
                    _rv.size,
                    hashlib.md5(_rv.tostring()).hexdigest() )
            fn = "func_results_ref_%s.cpkl" % (name)
            if not os.path.exists(fn):
                print("Answers need to be created with --answer-reference .")
                return False
            with open(fn, "rb") as f:
                ref = cPickle.load(f)
            print("Sizes: %s (%s, %s)" % (vals[4] == ref[4], vals[4], ref[4]))
            assert_allclose(vals[0], ref[0], 1e-8, err_msg="min")
            assert_allclose(vals[1], ref[1], 1e-8, err_msg="max")
            assert_allclose(vals[2], ref[2], 1e-8, err_msg="std")
            assert_allclose(vals[3], ref[3], 1e-8, err_msg="sum")
            assert_equal(vals[4], ref[4])
            print("Hashes equal: %s" % (vals[-1] == ref[-1]))
            return rv
        return _func
    return compare_results(func)

def periodicity_cases(ds):
    # This is a generator that yields things near the corners.  It's good for
    # getting different places to check periodicity.
    yield (ds.domain_left_edge + ds.domain_right_edge)/2.0
    dx = ds.domain_width / ds.domain_dimensions
    # We start one dx in, and only go to one in as well.
    for i in (1, ds.domain_dimensions[0] - 2):
        for j in (1, ds.domain_dimensions[1] - 2):
            for k in (1, ds.domain_dimensions[2] - 2):
                center = dx * np.array([i,j,k]) + ds.domain_left_edge
                yield center

def run_nose(verbose=False, run_answer_tests=False, answer_big_data=False,
             call_pdb=False, module=None):
    from yt.utilities.on_demand_imports import _nose
    import sys
    from yt.utilities.logger import ytLogger as mylog
    orig_level = mylog.getEffectiveLevel()
    mylog.setLevel(50)
    nose_argv = sys.argv
    nose_argv += ['--exclude=answer_testing','--detailed-errors', '--exe']
    if call_pdb:
        nose_argv += ["--pdb", "--pdb-failures"]
    if verbose:
        nose_argv.append('-v')
    if run_answer_tests:
        nose_argv.append('--with-answer-testing')
    if answer_big_data:
        nose_argv.append('--answer-big-data')
    if module:
        nose_argv.append(module)
    initial_dir = os.getcwd()
    yt_file = os.path.abspath(__file__)
    yt_dir = os.path.dirname(yt_file)
    if os.path.samefile(os.path.dirname(yt_dir), initial_dir):
        # Provide a nice error message to work around nose bug
        # see https://github.com/nose-devs/nose/issues/701
        raise RuntimeError(
            """
    The yt.run_nose function does not work correctly when invoked in
    the same directory as the installed yt package. Try starting
    a python session in a different directory before invoking yt.run_nose
    again. Alternatively, you can also run the "nosetests" executable in
    the current directory like so:

        $ nosetests
            """
            )
    os.chdir(yt_dir)
    try:
        _nose.run(argv=nose_argv)
    finally:
        os.chdir(initial_dir)
        mylog.setLevel(orig_level)

def assert_allclose_units(actual, desired, rtol=1e-7, atol=0, **kwargs):
    """Raise an error if two objects are not equal up to desired tolerance

    This is a wrapper for :func:`numpy.testing.assert_allclose` that also
    verifies unit consistency

    Parameters
    ----------
    actual : array-like
        Array obtained (possibly with attached units)
    desired : array-like
        Array to compare with (possibly with attached units)
    rtol : float, optional
        Relative tolerance, defaults to 1e-7
    atol : float or quantity, optional
        Absolute tolerance. If units are attached, they must be consistent
        with the units of ``actual`` and ``desired``. If no units are attached,
        assumes the same units as ``desired``. Defaults to zero.

    Notes
    -----
    Also accepts additional keyword arguments accepted by
    :func:`numpy.testing.assert_allclose`, see the documentation of that
    function for details.

    """
    # Create a copy to ensure this function does not alter input arrays
    act = YTArray(actual)
    des = YTArray(desired)

    try:
        des = des.in_units(act.units)
    except YTUnitOperationError:
        raise AssertionError("Units of actual (%s) and desired (%s) do not have "
                             "equivalent dimensions" % (act.units, des.units))

    rt = YTArray(rtol)
    if not rt.units.is_dimensionless:
        raise AssertionError("Units of rtol (%s) are not "
                             "dimensionless" % rt.units)

    if not isinstance(atol, YTArray):
        at = YTQuantity(atol, des.units)

    try:
        at = at.in_units(act.units)
    except YTUnitOperationError:
        raise AssertionError("Units of atol (%s) and actual (%s) do not have "
                             "equivalent dimensions" % (at.units, act.units))

    # units have been validated, so we strip units before calling numpy
    # to avoid spurious errors
    act = act.value
    des = des.value
    rt = rt.value
    at = at.value

    return assert_allclose(act, des, rt, at, **kwargs)

def assert_fname(fname):
    """Function that checks file type using libmagic"""
    if fname is None:
        return

    with open(fname, 'rb') as fimg:
        data = fimg.read()
    image_type = ''

    # see http://www.w3.org/TR/PNG/#5PNG-file-signature
    if data.startswith(b'\211PNG\r\n\032\n'):
        image_type = '.png'
    # see http://www.mathguide.de/info/tools/media-types/image/jpeg
    elif data.startswith(b'\377\330'):
        image_type = '.jpeg'
    elif data.startswith(b'%!PS-Adobe'):
        data_str = data.decode("utf-8", "ignore")
        if 'EPSF' in data_str[:data_str.index('\n')]:
            image_type = '.eps'
        else:
            image_type = '.ps'
    elif data.startswith(b'%PDF'):
        image_type = '.pdf'

    return image_type == os.path.splitext(fname)[1]




"""
Enzo-specific IO functions



"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import numpy as np

from yt.utilities.io_handler import \
    BaseIOHandler
from yt.utilities.logger import ytLogger as mylog
from yt.utilities.lib.geometry_utils import compute_morton
from yt.utilities.exceptions import YTDomainOverflow

class IOHandlerStream(BaseIOHandler):

    _dataset_type = "stream"
    _vector_fields = ("particle_velocity", "particle_position")

    def __init__(self, ds):
        self.fields = ds.stream_handler.fields
        self.field_units = ds.stream_handler.field_units
        super(IOHandlerStream, self).__init__(ds)

    def _read_data_set(self, grid, field):
        # This is where we implement processor-locking
        #if grid.id not in self.grids_in_memory:
        #    mylog.error("Was asked for %s but I have %s", grid.id, self.grids_in_memory.keys())
        #    raise KeyError
        tr = self.fields[grid.id][field]
        # If it's particles, we copy.
        if len(tr.shape) == 1: return tr.copy()
        # New in-place unit conversion breaks if we don't copy first
        return tr

    def _read_fluid_selection(self, chunks, selector, fields, size):
        chunks = list(chunks)
        if any((ftype not in self.ds.fluid_types for ftype, fname in fields)):
            raise NotImplementedError
        rv = {}
        for field in fields:
            rv[field] = self.ds.arr(np.empty(size, dtype="float64"))

        ng = sum(len(c.objs) for c in chunks)
        mylog.debug("Reading %s cells of %s fields in %s blocks",
                    size, [f2 for f1, f2 in fields], ng)
        for field in fields:
            ftype, fname = field
            ind = 0
            for chunk in chunks:
                for g in chunk.objs:
                    ds = self.fields[g.id][ftype, fname]
                    ind += g.select(selector, ds, rv[field], ind) # caches
        return rv

    def _read_particle_coords(self, chunks, ptf):
        chunks = list(chunks)
        for chunk in chunks:
            for g in chunk.objs:
                if g.NumberOfParticles == 0: continue
                gf = self.fields[g.id]
                for ptype, field_list in sorted(ptf.items()):
                    if (ptype, "particle_position") in gf:
                        x, y, z = gf[ptype, "particle_position"].T
                    else:
                        x, y, z = (gf[ptype, "particle_position_%s" % ax] for
                                   ax in 'xyz')
                    yield ptype, (x, y, z)

    def _read_particle_fields(self, chunks, ptf, selector):
        chunks = list(chunks)
        for chunk in chunks:
            for g in chunk.objs:
                if g.NumberOfParticles == 0: continue
                gf = self.fields[g.id]
                for ptype, field_list in sorted(ptf.items()):
                    if (ptype, "particle_position") in gf:
                        x, y, z = gf[ptype, "particle_position"].T
                    else:
                        x, y, z = (gf[ptype, "particle_position_%s" % ax] for
                                   ax in 'xyz')
                    mask = selector.select_points(x, y, z, 0.0)
                    if mask is None: continue
                    for field in field_list:
                        data = np.asarray(gf[ptype, field])
                        yield (ptype, field), data[mask]

    @property
    def _read_exception(self):
        return KeyError

class StreamParticleIOHandler(BaseIOHandler):

    _vector_fields = ("particle_position", "particle_velocity")
    _dataset_type = "stream_particles"
    _vector_fields = ("particle_velocity", "particle_position")

    def __init__(self, ds):
        self.fields = ds.stream_handler.fields
        super(StreamParticleIOHandler, self).__init__(ds)

    def _read_particle_coords(self, chunks, ptf):
        chunks = list(chunks)
        data_files = set([])
        for chunk in chunks:
            for obj in chunk.objs:
                data_files.update(obj.data_files)
        for data_file in sorted(data_files):
            f = self.fields[data_file.filename]
            # This double-reads
            for ptype, field_list in sorted(ptf.items()):
                yield ptype, (f[ptype, "particle_position_x"],
                              f[ptype, "particle_position_y"],
                              f[ptype, "particle_position_z"])
            
    def __count_particles_chunks(self, chunks, ptf, selector):
        # DISABLED
        # I have left this in here, but disabled, because of two competing
        # problems:
        #   * The IndexedOctreeSubsetSelector currently selects *all* particles
        #   * Slicing a deposited field thus throws an error, since the octree
        #     estimate fails.
        #   * BUT, it provides considerable speedup in some situations for
        #     stream datasets.
        # So, pending its re-enabling, we'll leave it here.
        # 
        # This is allowed to over-estimate.  We probably *will*, too, because
        # we're going to count *all* of the particles, not just individual
        # types.
        count = 0
        psize = {}
        for chunk in chunks:
            for obj in chunk.objs:
                count += selector.count_octs(obj.oct_handler, obj.domain_id)
        for ptype in ptf:
            psize[ptype] = self.ds.n_ref * count
        return psize

    def _read_particle_fields(self, chunks, ptf, selector):
        data_files = set([])
        for chunk in chunks:
            for obj in chunk.objs:
                data_files.update(obj.data_files)
        for data_file in sorted(data_files):
            f = self.fields[data_file.filename]
            for ptype, field_list in sorted(ptf.items()):
                if (ptype, "particle_position") in f:
                    x = f[ptype, "particle_position"][:,0]
                    y = f[ptype, "particle_position"][:,1]
                    z = f[ptype, "particle_position"][:,2]
                else:
                    x, y, z = (f[ptype, "particle_position_%s" % ax]
                               for ax in 'xyz')
                mask = selector.select_points(x, y, z, 0.0)
                if mask is None: continue
                for field in field_list:
                    data = f[ptype, field][mask]
                    yield (ptype, field), data

    def _initialize_index(self, data_file, regions):
        # self.fields[g.id][fname] is the pattern here
        morton = []
        for ptype in self.ds.particle_types_raw:
            try:
                pos = np.column_stack(self.fields[data_file.filename][
                    (ptype, "particle_position_%s" % ax)] for ax in 'xyz')
            except KeyError:
                pos = self.fields[data_file.filename][ptype, "particle_position"]
            if np.any(pos.min(axis=0) < data_file.ds.domain_left_edge) or \
               np.any(pos.max(axis=0) > data_file.ds.domain_right_edge):
                raise YTDomainOverflow(pos.min(axis=0), pos.max(axis=0),
                                       data_file.ds.domain_left_edge,
                                       data_file.ds.domain_right_edge)
            regions.add_data_file(pos, data_file.file_id)
            morton.append(compute_morton(
                    pos[:,0], pos[:,1], pos[:,2],
                    data_file.ds.domain_left_edge,
                    data_file.ds.domain_right_edge))
        return np.concatenate(morton)

    def _count_particles(self, data_file):
        pcount = {}
        for ptype in self.ds.particle_types_raw:
            d = self.fields[data_file.filename]
            try:
                pcount[ptype] = d[ptype, "particle_position_x"].size
            except KeyError:
                pcount[ptype] = d[ptype, "particle_position"].shape[0]
        return pcount

    def _identify_fields(self, data_file):
        return self.fields[data_file.filename].keys(), {}

class IOHandlerStreamHexahedral(BaseIOHandler):
    _dataset_type = "stream_hexahedral"
    _vector_fields = ("particle_velocity", "particle_position")

    def __init__(self, ds):
        self.fields = ds.stream_handler.fields
        super(IOHandlerStreamHexahedral, self).__init__(ds)

    def _read_fluid_selection(self, chunks, selector, fields, size):
        chunks = list(chunks)
        assert(len(chunks) == 1)
        chunk = chunks[0]
        rv = {}
        for field in fields:
            ftype, fname = field
            rv[field] = np.empty(size, dtype="float64")
        ngrids = sum(len(chunk.objs) for chunk in chunks)
        mylog.debug("Reading %s cells of %s fields in %s blocks",
                    size, [fn for ft, fn in fields], ngrids)
        for field in fields:
            ind = 0
            ftype, fname = field
            for chunk in chunks:
                for g in chunk.objs:
                    ds = self.fields[g.mesh_id].get(field, None)
                    if ds is None:
                        ds = self.fields[g.mesh_id][fname]
                    ind += g.select(selector, ds, rv[field], ind) # caches
        return rv

class IOHandlerStreamOctree(BaseIOHandler):
    _dataset_type = "stream_octree"
    _vector_fields = ("particle_velocity", "particle_position")

    def __init__(self, ds):
        self.fields = ds.stream_handler.fields
        super(IOHandlerStreamOctree, self).__init__(ds)

    def _read_fluid_selection(self, chunks, selector, fields, size):
        rv = {}
        ind = 0
        chunks = list(chunks)
        assert(len(chunks) == 1)
        for chunk in chunks:
            assert(len(chunk.objs) == 1)
            for subset in chunk.objs:
                field_vals = {}
                for field in fields:
                    field_vals[field] = self.fields[
                        subset.domain_id - subset._domain_offset][field]
                subset.fill(field_vals, rv, selector, ind)
        return rv


class IOHandlerStreamUnstructured(BaseIOHandler):
    _dataset_type = "stream_unstructured"

    def __init__(self, ds):
        self.fields = ds.stream_handler.fields
        super(IOHandlerStreamUnstructured, self).__init__(ds)

    def _read_fluid_selection(self, chunks, selector, fields, size):
        chunks = list(chunks)
        rv = {}
        for field in fields:
            ftype, fname = field
            if ftype == "all":
                ci = np.concatenate([mesh.connectivity_indices
                                     for mesh in self.ds.index.mesh_union])
            else:
                mesh_id = int(ftype[-1]) - 1
                m = self.ds.index.meshes[mesh_id]
                ci = m.connectivity_indices
            num_elem = ci.shape[0]
            if fname in self.ds._node_fields:
                nodes_per_element = ci.shape[1]
                rv[field] = np.empty((num_elem, nodes_per_element), dtype="float64")
            else:
                rv[field] = np.empty(num_elem, dtype="float64")
        for field in fields:
            ind = 0
            ftype, fname = field
            if ftype == "all":
                objs = [mesh for mesh in self.ds.index.mesh_union]
            else:
                mesh_ids = [int(ftype[-1])]
                chunk = chunks[mesh_ids[0] - 1]
                objs = chunk.objs
            for g in objs:
                ds = self.fields[g.mesh_id].get(field, None)
                if ds is None:
                    f = ('connect%d' % (g.mesh_id + 1), fname)
                    ds = self.fields[g.mesh_id][f]
                ind += g.select(selector, ds, rv[field], ind) # caches
        return rv




"""
API for yt.frontends.gdf



"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------
from .data_structures import \
      GDFGrid, \
      GDFHierarchy, \
      GDFDataset

from .fields import \
      GDFFieldInfo
add_gdf_field = GDFFieldInfo.add_field

from .io import \
      IOHandlerGDFHDF5

from . import tests



"""
Test for Volume Rendering Lenses.
"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import os
import tempfile
import shutil
from yt.testing import fake_random_ds
from yt.visualization.volume_rendering.api import Scene, VolumeSource
import numpy as np
from unittest import TestCase

def setup():
    """Test specific setup."""
    from yt.config import ytcfg
    ytcfg["yt", "__withintesting"] = "True"


class LensTest(TestCase):
    # This toggles using a temporary directory. Turn off to examine images.
    use_tmpdir = True

    def setUp(self):
        if self.use_tmpdir:
            self.curdir = os.getcwd()
            # Perform I/O in safe place instead of yt main dir
            self.tmpdir = tempfile.mkdtemp()
            os.chdir(self.tmpdir)
        else:
            self.curdir, self.tmpdir = None, None

        self.field = ("gas", "density")
        self.ds = fake_random_ds(32, fields=self.field)
        self.ds.index

    def tearDown(self):
        if self.use_tmpdir:
            os.chdir(self.curdir)
            shutil.rmtree(self.tmpdir)

    def test_perspective_lens(self):
        sc = Scene()
        cam = sc.add_camera(self.ds, lens_type='perspective')
        cam.position = self.ds.arr(np.array([1.0, 1.0, 1.0]), 'code_length')
        vol = VolumeSource(self.ds, field=self.field)
        tf = vol.transfer_function
        tf.grey_opacity = True
        sc.add_source(vol)
        sc.render()
        sc.save('test_perspective_%s.png' % self.field[1], sigma_clip=6.0)

    def test_stereoperspective_lens(self):
        sc = Scene()
        cam = sc.add_camera(self.ds, lens_type='stereo-perspective')
        cam.resolution = [1024, 512]
        cam.position = self.ds.arr(np.array([0.7, 0.7, 0.7]), 'code_length')
        vol = VolumeSource(self.ds, field=self.field)
        tf = vol.transfer_function
        tf.grey_opacity = True
        sc.add_source(vol)
        sc.render()
        sc.save('test_stereoperspective_%s.png' % self.field[1], sigma_clip=6.0)

    def test_fisheye_lens(self):
        dd = self.ds.sphere(self.ds.domain_center,
                            self.ds.domain_width[0] / 10)
        sc = Scene()
        cam = sc.add_camera(dd, lens_type='fisheye')
        cam.lens.fov = 360.0
        cam.set_width(self.ds.domain_width)
        v, c = self.ds.find_max('density')
        cam.set_position(c-0.0005*self.ds.domain_width)
        vol = VolumeSource(dd, field=self.field)
        tf = vol.transfer_function
        tf.grey_opacity = True
        sc.add_source(vol)
        sc.render()
        sc.save('test_fisheye_%s.png' % self.field[1], sigma_clip=6.0)

    def test_plane_lens(self):
        dd = self.ds.sphere(self.ds.domain_center,
                            self.ds.domain_width[0] / 10)
        sc = Scene()
        cam = sc.add_camera(dd, lens_type='plane-parallel')
        cam.set_width(self.ds.domain_width*1e-2)
        v, c = self.ds.find_max('density')
        vol = VolumeSource(dd, field=self.field)
        tf = vol.transfer_function
        tf.grey_opacity = True
        sc.add_source(vol)
        sc.render()
        sc.save('test_plane_%s.png' % self.field[1], sigma_clip=6.0)

    def test_spherical_lens(self):
        sc = Scene()
        cam = sc.add_camera(self.ds, lens_type='spherical')
        cam.resolution = [512, 256]
        cam.position = self.ds.arr(np.array([0.6, 0.5, 0.5]), 'code_length')
        vol = VolumeSource(self.ds, field=self.field)
        tf = vol.transfer_function
        tf.grey_opacity = True
        sc.add_source(vol)
        sc.render()
        sc.save('test_spherical_%s.png' % self.field[1], sigma_clip=6.0)

    def test_stereospherical_lens(self):
        w = (self.ds.domain_width).in_units('code_length')
        w = self.ds.arr(w, 'code_length')
        sc = Scene()
        cam = sc.add_camera(self.ds, lens_type='stereo-spherical')
        cam.resolution = [512, 512]
        cam.position = self.ds.arr(np.array([0.6, 0.5, 0.5]), 'code_length')
        vol = VolumeSource(self.ds, field=self.field)
        tf = vol.transfer_function
        tf.grey_opacity = True
        sc.add_source(vol)
        sc.render()
        sc.save('test_stereospherical_%s.png' % self.field[1], sigma_clip=6.0)



"""
Unit test FITS image creation in yt.



"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import tempfile
import os
import shutil
from yt.testing import fake_random_ds, requires_module
from yt.convenience import load
from numpy.testing import \
    assert_equal
from yt.visualization.fits_image import \
    FITSImageData, FITSProjection, \
    FITSSlice, FITSOffAxisSlice, \
    FITSOffAxisProjection, \
    assert_same_wcs
from yt.visualization.volume_rendering.off_axis_projection import \
    off_axis_projection


@requires_module("astropy")
def test_fits_image():
    tmpdir = tempfile.mkdtemp()
    curdir = os.getcwd()
    os.chdir(tmpdir)

    fields = ("density", "temperature")
    units = ('g/cm**3', 'K',)
    ds = fake_random_ds(64, fields=fields, units=units, nprocs=16,
                        length_unit=100.0)

    prj = ds.proj("density", 2)
    prj_frb = prj.to_frb((0.5, "unitary"), 128)

    fid1 = FITSImageData(prj_frb, fields=["density","temperature"], units="cm")
    fits_prj = FITSProjection(ds, "z", [ds.fields.gas.density,"temperature"], image_res=128,
                              width=(0.5,"unitary"))

    assert_equal(fid1["density"].data, fits_prj["density"].data)
    assert_equal(fid1["temperature"].data, fits_prj["temperature"].data)

    fid1.writeto("fid1.fits", clobber=True)
    new_fid1 = FITSImageData.from_file("fid1.fits")

    assert_equal(fid1["density"].data, new_fid1["density"].data)
    assert_equal(fid1["temperature"].data, new_fid1["temperature"].data)

    ds2 = load("fid1.fits")
    ds2.index

    assert ("fits","density") in ds2.field_list
    assert ("fits","temperature") in ds2.field_list

    dw_cm = ds2.domain_width.in_units("cm")

    assert dw_cm[0].v == 50.
    assert dw_cm[1].v == 50.

    slc = ds.slice(2, 0.5)
    slc_frb = slc.to_frb((0.5, "unitary"), 128)

    fid2 = FITSImageData(slc_frb, fields=["density","temperature"], units="cm")
    fits_slc = FITSSlice(ds, "z", ["density",("gas","temperature")], image_res=128,
                         width=(0.5,"unitary"))

    assert_equal(fid2["density"].data, fits_slc["density"].data)
    assert_equal(fid2["temperature"].data, fits_slc["temperature"].data)

    dens_img = fid2.pop("density")
    temp_img = fid2.pop("temperature")

    # This already has some assertions in it, so we don't need to do anything
    # with it other than just make one
    FITSImageData.from_images([dens_img, temp_img])

    cut = ds.cutting([0.1, 0.2, -0.9], [0.5, 0.42, 0.6])
    cut_frb = cut.to_frb((0.5, "unitary"), 128)

    fid3 = FITSImageData(cut_frb, fields=[("gas","density"), ds.fields.gas.temperature], units="cm")
    fits_cut = FITSOffAxisSlice(ds, [0.1, 0.2, -0.9], ["density","temperature"],
                                image_res=128, center=[0.5, 0.42, 0.6],
                                width=(0.5,"unitary"))

    assert_equal(fid3["density"].data, fits_cut["density"].data)
    assert_equal(fid3["temperature"].data, fits_cut["temperature"].data)

    fid3.create_sky_wcs([30.,45.], (1.0,"arcsec/kpc"))
    fid3.writeto("fid3.fits", clobber=True)
    new_fid3 = FITSImageData.from_file("fid3.fits")
    assert_same_wcs(fid3.wcs, new_fid3.wcs)
    assert new_fid3.wcs.wcs.cunit[0] == "deg"
    assert new_fid3.wcs.wcs.cunit[1] == "deg"
    assert new_fid3.wcs.wcs.ctype[0] == "RA---TAN"
    assert new_fid3.wcs.wcs.ctype[1] == "DEC--TAN"

    buf = off_axis_projection(ds, ds.domain_center, [0.1, 0.2, -0.9],
                              0.5, 128, "density").swapaxes(0, 1)
    fid4 = FITSImageData(buf, fields="density", width=100.0)
    fits_oap = FITSOffAxisProjection(ds, [0.1, 0.2, -0.9], "density",
                                     width=(0.5, "unitary"), image_res=128,
                                     depth=(0.5, "unitary"))

    assert_equal(fid4["density"].data, fits_oap["density"].data)

    fid4.create_sky_wcs([30., 45.], (1.0, "arcsec/kpc"), replace_old_wcs=False)
    assert fid4.wcs.wcs.cunit[0] == "cm"
    assert fid4.wcs.wcs.cunit[1] == "cm"
    assert fid4.wcs.wcs.ctype[0] == "linear"
    assert fid4.wcs.wcs.ctype[1] == "linear"
    assert fid4.wcsa.wcs.cunit[0] == "deg"
    assert fid4.wcsa.wcs.cunit[1] == "deg"
    assert fid4.wcsa.wcs.ctype[0] == "RA---TAN"
    assert fid4.wcsa.wcs.ctype[1] == "DEC--TAN"

    cvg = ds.covering_grid(ds.index.max_level, [0.25, 0.25, 0.25],
                           [32, 32, 32], fields=["density", "temperature"])
    fid5 = FITSImageData(cvg, fields=["density", "temperature"])
    assert fid5.dimensionality == 3

    fid5.update_header("density", "time", 0.1)
    fid5.update_header("all", "units", "cgs")

    assert fid5["density"].header["time"] == 0.1
    assert fid5["temperature"].header["units"] == "cgs"
    assert fid5["density"].header["units"] == "cgs"

    os.chdir(curdir)
    shutil.rmtree(tmpdir)



"""
Volume rendering

"""

#-----------------------------------------------------------------------------
# Copyright (c) 2014, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------


from .scene import Scene
from .render_source import VolumeSource
from .transfer_functions import ProjectionTransferFunction
from .utils import data_source_or_all
from yt.funcs import mylog, iterable
from yt.utilities.lib.partitioned_grid import \
    PartitionedGrid
from yt.data_objects.api import ImageArray
import numpy as np


def off_axis_projection(data_source, center, normal_vector,
                        width, resolution, item,
                        weight=None, volume=None,
                        no_ghost=False, interpolated=False,
                        north_vector=None, num_threads=1, method='integrate'):
    r"""Project through a dataset, off-axis, and return the image plane.

    This function will accept the necessary items to integrate through a volume
    at an arbitrary angle and return the integrated field of view to the user.
    Note that if a weight is supplied, it will multiply the pre-interpolated
    values together, then create cell-centered values, then interpolate within
    the cell to conduct the integration.

    Parameters
    ----------
    data_source : ~yt.data_objects.static_output.Dataset or ~yt.data_objects.data_containers.YTSelectionDataContainer
        This is the dataset or data object to volume render.
    center : array_like
        The current 'center' of the view port -- the focal point for the
        camera.
    normal_vector : array_like
        The vector between the camera position and the center.
    width : float or list of floats
        The current width of the image.  If a single float, the volume is
        cubical, but if not, it is left/right, top/bottom, front/back
    resolution : int or list of ints
        The number of pixels in each direction.
    item: string
        The field to project through the volume
    weight : optional, default None
        If supplied, the field will be pre-multiplied by this, then divided by
        the integrated value of this field.  This returns an average rather
        than a sum.
    volume : `yt.extensions.volume_rendering.AMRKDTree`, optional
        The volume to ray cast through.  Can be specified for finer-grained
        control, but otherwise will be automatically generated.
    no_ghost: bool, optional
        Optimization option.  If True, homogenized bricks will
        extrapolate out from grid instead of interpolating from
        ghost zones that have to first be calculated.  This can
        lead to large speed improvements, but at a loss of
        accuracy/smoothness in resulting image.  The effects are
        less notable when the transfer function is smooth and
        broad. Default: True
    interpolated : optional, default False
        If True, the data is first interpolated to vertex-centered data,
        then tri-linearly interpolated along the ray. Not suggested for
        quantitative studies.
    north_vector : optional, array_like, default None
        A vector that, if specified, restricts the orientation such that the
        north vector dotted into the image plane points "up". Useful for rotations
    num_threads: integer, optional, default 1
        Use this many OpenMP threads during projection.
    method : string
        The method of projection.  Valid methods are:

        "integrate" with no weight_field specified : integrate the requested
        field along the line of sight.

        "integrate" with a weight_field specified : weight the requested
        field by the weighting field and integrate along the line of sight.

        "sum" : This method is the same as integrate, except that it does not
        multiply by a path length when performing the integration, and is
        just a straight summation of the field along the given axis. WARNING:
        This should only be used for uniform resolution grid datasets, as other
        datasets may result in unphysical images.
        or camera movements.
    Returns
    -------
    image : array
        An (N,N) array of the final integrated values, in float64 form.

    Examples
    --------

    >>> image = off_axis_projection(ds, [0.5, 0.5, 0.5], [0.2,0.3,0.4],
    ...                             0.2, N, "temperature", "density")
    >>> write_image(np.log10(image), "offaxis.png")

    """

    if method not in ['integrate','sum']:
        raise NotImplementedError("Only 'integrate' or 'sum' methods are valid for off-axis-projections")

    if interpolated is True:
        raise NotImplementedError("Only interpolated=False methods are currently implemented for off-axis-projections")


    data_source = data_source_or_all(data_source)
    # Sanitize units
    if not hasattr(center, "units"):
        center = data_source.ds.arr(center, 'code_length')
    if not hasattr(width, "units"):
        width = data_source.ds.arr(width, 'code_length')
    sc = Scene()
    data_source.ds.index
    if item is None:
        field = data_source.ds.field_list[0]
        mylog.info('Setting default field to %s' % field.__repr__())

    funits = data_source.ds._get_field_info(item).units

    vol = VolumeSource(data_source, item)
    if weight is None:
        vol.set_field(item)
    else:
        # This is a temporary field, which we will remove at the end.
        weightfield = ("index", "temp_weightfield")
        def _make_wf(f, w):
            def temp_weightfield(a, b):
                tr = b[f].astype("float64") * b[w]
                return b.apply_units(tr, a.units)
            return temp_weightfield
        data_source.ds.field_info.add_field(weightfield, sampling_type="cell",
            function=_make_wf(item, weight))
        # Now we have to tell the dataset to add it and to calculate
        # its dependencies..
        deps, _ = data_source.ds.field_info.check_derived_fields([weightfield])
        data_source.ds.field_dependencies.update(deps)
        vol.set_field(weightfield)
        vol.set_weight_field(weight)
    ptf = ProjectionTransferFunction()
    vol.set_transfer_function(ptf)
    camera = sc.add_camera(data_source)
    camera.set_width(width)
    if not iterable(resolution):
        resolution = [resolution]*2
    camera.resolution = resolution
    if not iterable(width):
        width = data_source.ds.arr([width]*3)
    normal = np.array(normal_vector)
    normal = normal / np.linalg.norm(normal)

    camera.position = center - width[2]*normal
    camera.focus = center

    # If north_vector is None, we set the default here.
    # This is chosen so that if normal_vector is one of the
    # cartesian coordinate axes, the projection will match
    # the corresponding on-axis projection.
    if north_vector is None:
        vecs = np.identity(3)
        t = np.cross(vecs, normal).sum(axis=1)
        ax = t.argmax()
        east_vector = np.cross(vecs[ax, :], normal).ravel()
        north = np.cross(normal, east_vector).ravel()
    else:
        north = np.array(north_vector)
        north = north / np.linalg.norm(north)
    camera.switch_orientation(normal, north)

    sc.add_source(vol)

    vol.set_sampler(camera, interpolated=False)
    assert (vol.sampler is not None)

    fields = [vol.field]
    if vol.weight_field is not None:
        fields.append(vol.weight_field)

    mylog.debug("Casting rays")

    for i, (grid, mask) in enumerate(data_source.blocks):
        data = [(grid[f] * mask).astype("float64") for f in fields]
        pg = PartitionedGrid(
            grid.id, data,
            mask.astype('uint8'),
            grid.LeftEdge, grid.RightEdge, grid.ActiveDimensions.astype("int64"))
        grid.clear_data()
        vol.sampler(pg, num_threads = num_threads)

    image = vol.finalize_image(camera, vol.sampler.aimage)
    image = ImageArray(image, funits, registry=data_source.ds.unit_registry, info=image.info)

    if weight is not None:
        data_source.ds.field_info.pop(("index", "temp_weightfield"))

    if method == "integrate":
        if weight is None:
            dl = width[2].in_units(data_source.ds.unit_system["length"])
            image *= dl
        else:
            mask = image[:,:,1] == 0
            image[:,:,0] /= image[:,:,1]
            image[mask] = 0

    return image[:,:,0]



"""
The basic field info container resides here.  These classes, code specific and
universal, are the means by which we access fields across YT, both derived and
native.



"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import numpy as np

from .derived_field import \
    ValidateParameter

from .field_plugin_registry import \
    register_field_plugin

from .vector_operations import \
    create_magnitude_field

from yt.utilities.lib.geometry_utils import \
    obtain_rvec, obtain_rv_vec

def obtain_velocities(data, ftype="gas"):
    rv_vec = obtain_rv_vec(data)
    # We know that obtain_rv_vec will always access velocity_x
    rv_vec = data.ds.arr(rv_vec, input_units = data[ftype, "velocity_x"].units)
    return rv_vec

@register_field_plugin
def setup_angular_momentum(registry, ftype = "gas", slice_info = None):
    unit_system = registry.ds.unit_system
    def _specific_angular_momentum_x(field, data):
        xv, yv, zv = obtain_velocities(data, ftype)
        rv = obtain_rvec(data)
        rv = np.rollaxis(rv, 0, len(rv.shape))
        rv = data.ds.arr(rv, input_units = data["index", "x"].units)
        return yv * rv[...,2] - zv * rv[...,1]

    def _specific_angular_momentum_y(field, data):
        xv, yv, zv = obtain_velocities(data, ftype)
        rv = obtain_rvec(data)
        rv = np.rollaxis(rv, 0, len(rv.shape))
        rv = data.ds.arr(rv, input_units = data["index", "x"].units)
        return - (xv * rv[...,2] - zv * rv[...,0])

    def _specific_angular_momentum_z(field, data):
        xv, yv, zv = obtain_velocities(data, ftype)
        rv = obtain_rvec(data)
        rv = np.rollaxis(rv, 0, len(rv.shape))
        rv = data.ds.arr(rv, input_units = data["index", "x"].units)
        return xv * rv[...,1] - yv * rv[...,0]

    registry.add_field((ftype, "specific_angular_momentum_x"), sampling_type="cell", 
                        function=_specific_angular_momentum_x,
                        units=unit_system["specific_angular_momentum"],
                        validators=[ValidateParameter("center")])
    registry.add_field((ftype, "specific_angular_momentum_y"), sampling_type="cell", 
                        function=_specific_angular_momentum_y,
                        units=unit_system["specific_angular_momentum"],
                        validators=[ValidateParameter("center")])
    registry.add_field((ftype, "specific_angular_momentum_z"), sampling_type="cell", 
                        function=_specific_angular_momentum_z,
                        units=unit_system["specific_angular_momentum"],
                        validators=[ValidateParameter("center")])

    create_magnitude_field(registry, "specific_angular_momentum",
                           unit_system["specific_angular_momentum"], ftype=ftype)

    def _angular_momentum_x(field, data):
        return data[ftype, "cell_mass"] \
             * data[ftype, "specific_angular_momentum_x"]
    registry.add_field((ftype, "angular_momentum_x"), sampling_type="cell", 
                       function=_angular_momentum_x,
                       units=unit_system["angular_momentum"],
                       validators=[ValidateParameter('center')])

    def _angular_momentum_y(field, data):
        return data[ftype, "cell_mass"] \
             * data[ftype, "specific_angular_momentum_y"]
    registry.add_field((ftype, "angular_momentum_y"), sampling_type="cell", 
                       function=_angular_momentum_y,
                       units=unit_system["angular_momentum"],
                       validators=[ValidateParameter('center')])

    def _angular_momentum_z(field, data):
        return data[ftype, "cell_mass"] \
             * data[ftype, "specific_angular_momentum_z"]
    registry.add_field((ftype, "angular_momentum_z"), sampling_type="cell", 
                       function=_angular_momentum_z,
                       units=unit_system["angular_momentum"],
                       validators=[ValidateParameter('center')])

    create_magnitude_field(registry, "angular_momentum",
                           unit_system["angular_momentum"], ftype=ftype)
